(define (get-string i)
  (let  ((n (read-char i)))
    (cond
      ((eof-object? n) '())
      (else (cons n (get-string i))))))

(define (eval-file f)
  (call-with-input-file f
    (lambda (i)
      (let ((c (compile (scan&parse (list->string (get-string i)))
                        compiler-start-env)))
        ;;(eopl:printf "~a~n" (cexp->sexp c))
        (cont-eval-exp c
                       cont-eval-start-env
                       (done-cont))))))

;; --------------------------------- Syntax -----------------------------------

(define the-lexical-spec
  '((whitespace (whitespace)
                skip)
    (id ((or letter "!" "$" "%" "&" "*" "/" ":" "<" "=" ">" "?" "^" "_" "~")
         (arbno (or letter digit
                    "!" "$" "%" "&" "*" "/" ":" "<" "=" ">"
                    "?" "^" "_" "~" "+" "-" "." "@")))
        make-symbol)
    (id ("+") make-symbol)
    (id ("-") make-symbol)
    (id ("...") make-symbol)
    (number ((or "" "-" "+") digit (arbno digit) (arbno "#"))
            make-number)))

(define the-grammar
  '((expression (number)
                num-exp)
    (expression ("null")
                null-exp)
    (expression ("#f")
                false-exp)
    (expression ("void")
                void-exp)
    (expression ("(lambda" "(" (arbno id) ")" expression ")")
                lambda-exp)
    (expression (id)
                var-exp)
    (expression ("(let" "(" (arbno "(" id expression ")") ")" expression ")")
                let-exp)
    (expression ("(letrec" "(" (arbno "(" id expression ")") ")" expression ")")
                letrec-exp)
    (expression ("(" expression (arbno expression) ")")
                app-exp)
    (expression ("(if" expression expression expression ")")
                if-exp)
    (expression ("(begin" (arbno expression) ")")
                begin-exp)
    (expression ("(set!" id expression ")")
                set!-exp)
    (expression ("(and" (arbno expression) ")")
                and-exp)
    (expression ("(or" (arbno expression) ")")
                or-exp)
    (expression ("(not" (arbno expression) ")")
                not-exp)))

(sllgen:make-define-datatypes the-lexical-spec the-grammar)
(define scan&parse (sllgen:make-string-parser the-lexical-spec the-grammar))

(define-datatype cexpression cexpression?
  (val-cexp (v value?))
  (lambda-cexp (num-parms number?)
               (body cexpression?))
  (lexvar-cexp (dist number?)
               (offset number?))
  (app-cexp (fun cexpression?)
            (args vector?))
  (if-cexp (test-cexp cexpression?)
           (t_branch-cexp cexpression?)
           (f_branch-cexp cexpression?))
  (lexset!-cexp (dist number?)
                (offset number?)
                (setto cexpression?))
  (and-cexp (cexps vector?))
  (or-cexp (cexps vector?))
  (not-cexp (cexp cexpression?)))

(define (cexp->sexp e)
  (cases cexpression e
    (val-cexp (v) v)
    (lambda-cexp (np b) `(lambda ,np ,(cexp->sexp b)))
    (lexvar-cexp (d o) `(lexvar ,d ,o))
    (app-cexp (f a) `(,(cexp->sexp f)
                      ,@(map cexp->sexp (vector->list a))))
    (if-cexp (a b c) `(if ,(cexp->sexp a) ,(cexp->sexp b) ,(cexp->sexp c)))
    (lexset!-cexp (d o s) `(lexset! ,d ,o ,(cexp->sexp s)))
    (and-cexp (a) `(and ,@(map cexp->sexp (vector->list a))))
    (or-cexp (a) `(or ,@(map cexp->sexp (vector->list a))))
    (not-cexp (a) `(not ,(cexp->sexp a)))))

;; ----------------------------- Environments ---------------------------------
(define-datatype environment environment?
  (empty-env-record)
  (extended-env-record
    (syms vector?)
    (vec vector?)
    (env environment?)))

(define (extend-env syms vals env)
  (extended-env-record syms vals env))

(define (get-rib sym env i)
  (cases environment env
    (empty-env-record ()
      (eopl:error 'env-lookup "No binding for ~s" sym))
    (extended-env-record (syms vals env)
      (let ((position (env-find-position 0 sym syms)))
        (if (number? position)
            (list i position vals)
            (get-rib sym env (+ i 1)))))))

(define (env-find-position i sym syms)
  (cond
    ((>= i (vector-length syms)) #f)
    ((eqv? sym (vector-ref syms i)) i)
    (else (env-find-position (+ i 1) sym syms))))      
  
(define (compute-lexical-address sym env)
  (let ((rib (get-rib sym env 0)))
    (lexvar-cexp (car rib) (cadr rib))))

(define (apply-env-lex dist offset env)
  (cases environment env
    (extended-env-record (syms vals env)
      (if (= 0 dist)
          (vector-ref vals offset)
          (apply-env-lex (- dist 1) offset env)))
    (empty-env-record ()
      (eopl:error 'lexical-lookup))))

(define (update-env-lex! dist offset val env)
  (cases environment env
    (extended-env-record (syms vals env)
      (if (= 0 dist)
          (vector-set! vals offset val)
          (update-env-lex! (- dist 1) offset val env)))
    (empty-env-record ()
      (eopl:error 'lexical-lookup))))
   

;; ----------------------------- Compiler -------------------------------------
;; compile: expression * environment -> cexpression
(define (compile exp env)
  (cases expression exp
    (num-exp (num) (val-cexp (number-val num)))
    (null-exp () (val-cexp (null-val)))
    (false-exp () (val-cexp (false-val)))
    (void-exp () (val-cexp (void-val)))
    (lambda-exp (parms body)
      (lambda-cexp (length parms)
                   (compile body
                            (extend-env (list->vector parms) (vector) env))))
    (var-exp (var) (compute-lexical-address var env))
    (let-exp (ids bound-exps body)
      (compile (compile-let ids bound-exps body) env))
    (letrec-exp (ids bound-exps body)
      (compile (compile-letrec ids bound-exps body) env))
    (app-exp (func args)
      (app-cexp (compile func env)
                (list->vector (map (lambda (e) (compile e env)) args))))
    (if-exp (e1 e2 e3)
      (if-cexp (compile e1 env) (compile e2 env) (compile e3 env)))
    (begin-exp (exps)
      (cond
        ((null? exps) (eopl:error 'compile "Empty begin expression"))
        (else (compile (compile-begin exps) env))))
    (set!-exp (var exp)
      (cases cexpression (compute-lexical-address var env)
        (lexvar-cexp (a b)
          (lexset!-cexp a b (compile exp env)))
        (else
         (eopl:error 'compiler "compute-lexical-address is broken"))))
    (and-exp (exps)
      (cond
       ((null? exps)
	(val-cexp (number-val 0)))
       (else
	(and-cexp (list->vector (map (lambda (x) (compile x env)) exps))))))
    (or-exp (exps)
      (cond
       ((null? exps)
	(val-cexp (false-val)))
       (else
	(or-cexp (list->vector (map (lambda (x) (compile x env)) exps))))))
    (not-exp (exp)
      (not-cexp (compile exp env)))))

;; compile-let: listof-symbol * listof-expression * expression -> expression
(define (compile-let ids bound-exps body)
  (app-exp (lambda-exp ids body) bound-exps))

;; compile-letrec: listof-symbol * listof-expression * expression -> expression
(define (compile-letrec ids bound-exps body)
  (let-exp ids 
           (map (lambda (x) (void-exp)) ids)
           (begin-exp
            (append (map (lambda (id be) (set!-exp id be)) ids bound-exps)
                    (list body)))))

;; compile-begin: listof-expression -> expression
(define (compile-begin exps)
  (let ((split (split-list exps (- (length exps) 1) '())))
    (app-exp (lambda-exp (map (lambda (x) (string->symbol "12")) (car split))
                         (car (cdr split)))
             (car split))))

;; split-list: list * number * list -> (cons list list)
(define (split-list l n acc)
  (cond
    ((= 0 n) (cons (reverse acc) l))
    (else
     (split-list (cdr l) (- n 1) (cons (car l) acc)))))

(define compiler-start-env
  (extended-env-record (vector 'cons 'car 'cdr 'set-car! 'set-cdr!
			       'null? 'pair? '= '< '> '+ '- '*)
                       (vector)
                       (empty-env-record)))


;; ------------------------------ Runtime Values ------------------------------
(define-datatype value value?
  (number-val (num number?))
  (null-val)
  (false-val)
  (closure (num-parms number?)
           (body-exp cexpression?)
           (env environment?))
  (void-val)
  (primitive-func (num-parms number?)
                  (f procedure?))
  (pair-val (sub-parts vector?)))

(define (print-value v)
  (cases value v
    (number-val (n) (number->string n))
    (null-val () "null")
    (false-val () "#f")
    (closure (a b c) "closure")
    (void-val () "void")
    (primitive-func (a b) "primitive function")
    (pair-val (parts)
      (string-append "(cons " (print-value (vector-ref parts 0)) " "
                     (print-value (vector-ref parts 1)) ")"))))

;; --------------------------------- Continuations ----------------------------

(define-datatype continuation continuation?
  (done-cont)
  (narg-cont (args vector?)
             (env environment?)
             (cont continuation?))
  (farg-cont (fun value?)
             (index number?)
             (args vector?)
             (env environment?)
             (cont continuation?))
  (if-test-cont (t_branch cexpression?)
                (f_branch cexpression?)
                (env environment?)
                (cont continuation?))
  (lexset!-cont (dist number?)
                (offset number?)
                (env environment?)
                (cont continuation?))
  (and-cont (index number?)
            (exps vector?)
            (env environment?)
            (cont continuation?))
  (or-cont (index number?)
           (exps vector?)
           (env environment?)
           (cont continuation?))
  (not-cont (cont continuation?)))

(define (apply-cont cont val)
  (cases continuation cont
    (done-cont () (print-value val))
    (narg-cont (args env cont)
      (cond
        ((= 0 (vector-length args))
         (apply-cont (farg-cont val 0 args env cont) #f))
        (else
         (cont-eval-exp (vector-ref args 0)
                          env
                          (farg-cont val 1 (vector-copy args) env cont)))))
    (farg-cont (fun-val index args env cont)
      (if (> index 0)
          (vector-set! args (- index 1) val))
      (cond
        ((= (vector-length args) index)
         (cases value fun-val
           (closure (num-parms body clo-env)
             (cond
               ((= num-parms index)
                (cont-eval-exp body
                                 (extend-env (vector) args clo-env)
                                 cont))
               (else
                (string-append "Function expects " num-parms 
                               " args, given " index "."))))
           (primitive-func (num-args f)
             (cond
               ((= num-args index) (f args cont))
               (else
                (string-append "Function expects " num-args
                               " args, given " index "."))))
           (else "Attempt to apply non-function")))
        (else
         (cont-eval-exp (vector-ref args index)
                          env
                          (farg-cont fun-val (+ index 1) args env cont)))))
    (if-test-cont (t_branch f_branch env cont)
      (cases value val
        (false-val () (cont-eval-exp f_branch env cont))
        (else (cont-eval-exp t_branch env cont))))
    (lexset!-cont (dist offset env cont)
      (begin
        (update-env-lex! dist offset val env)
        (apply-cont cont (void-val))))
    (and-cont (index exps env cont)
      (cond
        ((= index (vector-length exps)) 
	 (apply-cont cont val))
        (else
         (cases value val
           (false-val () (apply-cont cont val))
           (else (cont-eval-exp (vector-ref exps index)
                                  env
                                  (and-cont (+ index 1) exps env cont)))))))
    (or-cont (index exps env cont)
      (cond
        ((= index (vector-length exps))
	 (apply-cont cont val))
        (else
         (cases value val
           (false-val () (cont-eval-exp (vector-ref exps index)
                                          env
                                          (and-cont (+ index 1) exps env cont)))
           (else (apply-cont cont val))))))
    (not-cont (cont)
      (cases value val
        (false-val () (apply-cont cont (number-val 0)))
        (else (apply-cont cont (false-val)))))))
        
;; ---------------------- Continuation Run-time Support ------------------------
(define (cont-num-op op)
  (primitive-func
   2 
   (lambda (arg cont)
     (cases value (vector-ref arg 0)
       (number-val (num1)
         (cases value (vector-ref arg 1)
           (number-val (num2)
             (apply-cont cont (number-val (op num1 num2))))
           (else
            (string-append "Numeric op given non-numeric arg "
                           (print-value (vector-ref arg 1))))))
       (else
        (string-append "Numeric op given non-numeric arg "
                       (print-value (vector-ref arg 0))))))))

(define (cont-comp-op op)
  (primitive-func
   2 
   (lambda (arg cont)
     (cases value (vector-ref arg 0)
       (number-val (num1)
         (cases value (vector-ref arg 1)
           (number-val (num2)
             (apply-cont cont
                         (if (op num1 num2)
                             (number-val 0)
                             (false-val))))
           (else
            (string-append "Numeric op given non-numeric arg "
                           (print-value (vector-ref arg 1))))))
       (else
        (string-append "Numeric op given non-numeric arg "
                       (print-value (vector-ref arg 0))))))))

;; 'cons 'car 'cdr 'set-car! 'set-cdr! 'null? 'pair? '= '< '> '+ '- '*
(define cont-eval-start-env
  (extended-env-record
   (vector)
   (vector
    (primitive-func 2 (lambda (arg cont) (apply-cont cont (pair-val arg))))
    (primitive-func 1 (lambda (arg cont)
                        (cases value (vector-ref arg 0)
                          (pair-val (sub-parts)
                            (apply-cont cont (vector-ref sub-parts 0)))
                          (else
                           (string-append
                            "car of " (print-value (vector-ref arg 0)))))))
    (primitive-func 1 (lambda (arg cont)
                        (cases value (vector-ref arg 0)
                          (pair-val (sub-parts)
                            (apply-cont cont (vector-ref sub-parts 1)))
                          (else
                           (string-append
                            "car of " (print-value (vector-ref arg 0)))))))
    (primitive-func 2 (lambda (arg cont)
                        (cases value (vector-ref arg 0)
                          (pair-val (sub-parts)
                            (begin
                              (vector-set!  sub-parts 0 (vector-ref arg 1))
                              (apply-cont cont (void-val))))
                          (else
                           (string-append
                            "set-car! of " (print-value (vector-ref arg 0)))))))
    (primitive-func 2 (lambda (arg cont)
                        (cases value (vector-ref arg 0)
                          (pair-val (sub-parts)
                            (begin
                              (vector-set! sub-parts 1 (vector-ref arg 1))
                              (apply-cont cont (void-val))))
                          (else
                           (string-append
                            "set-cdr! of " (print-value (vector-ref arg 0)))))))
    (primitive-func 1 (lambda (arg cont)
                        (apply-cont cont
                                    (cases value (vector-ref arg 0)
                                      (null-val () (vector-ref arg 0))
                                      (else (false-val))))))
    (primitive-func 1 (lambda (arg cont)
                        (apply-cont cont
                                    (cases value (vector-ref arg 0)
                                      (pair-val (a) (vector-ref arg 0))
                                      (else (false-val))))))
    (cont-comp-op =)
    (cont-comp-op <)
    (cont-comp-op >)
    (cont-num-op +)
    (cont-num-op -)
    (cont-num-op *))
   (empty-env-record)))



;; ------------------------ Continuation based interpreter ---------------------
(define (cont-eval-exp exp env cont)
  (cases cexpression exp
    (val-cexp (v) (apply-cont cont v))
    (lambda-cexp (num-parms body)
      (apply-cont cont (closure num-parms body env)))
    (lexvar-cexp (dist offset)
      (apply-cont cont (apply-env-lex dist offset env)))
    (app-cexp (fun-exp arg-exps)
      (cont-eval-exp fun-exp env (narg-cont arg-exps env cont)))
    (if-cexp (test t_branch f_branch)
      (cont-eval-exp test env (if-test-cont t_branch f_branch env cont)))
    (lexset!-cexp (dist offset expr)
      (cont-eval-exp expr env (lexset!-cont dist offset env cont)))
    (and-cexp (exps)
      (cont-eval-exp (vector-ref exps 0)
		     env
		     (and-cont 1 exps env cont)))
    (or-cexp (exps)
      (cont-eval-exp (vector-ref exps 0)
		     env
		     (or-cont 1 exps env cont)))
    (not-cexp (exp)
      (cont-eval-exp exp env (not-cont cont)))))

(define vector-copy
  (letrec ((vcopy-helper
            (lambda (v-old v-new idx)
              (cond
                ((= 0 idx) v-new)
                (else (vector-set! v-new idx (vector-ref v-old idx))
                      (vcopy-helper v-old v-new (- idx 1)))))))
    (lambda (v)
      (vcopy-helper v (make-vector (vector-length v))
                    (- (vector-length v) 1)))))